# Copyright 2010-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["env_update"]

import errno
import glob
import stat
import subprocess
import time

import portage
from portage import os, _encodings, _unicode_decode, _unicode_encode
from portage.checksum import prelink_capable
from portage.data import ostype
from portage.exception import ParseError
from portage.localization import _
from portage.process import find_binary
from portage.util import (
    atomic_ofstream,
    ensure_dirs,
    getconfig,
    normalize_path,
    writemsg,
)
from portage.util.listdir import listdir
from portage.dbapi.vartree import vartree
from portage.package.ebuild.config import config


def env_update(
    makelinks=1,
    target_root=None,
    prev_mtimes=None,
    contents=None,
    env=None,
    writemsg_level=None,
    vardbapi=None,
):
    """
    Parse /etc/env.d and use it to generate /etc/profile.env, csh.env,
    ld.so.conf, and prelink.conf. Finally, run ldconfig. When ldconfig is
    called, its -X option will be used in order to avoid potential
    interference with installed soname symlinks that are required for
    correct operation of FEATURES=preserve-libs for downgrade operations.
    It's not necessary for ldconfig to create soname symlinks, since
    portage will use NEEDED.ELF.2 data to automatically create them
    after src_install if they happen to be missing.
    @param makelinks: True if ldconfig should be called, False otherwise
    @param target_root: root that is passed to the ldconfig -r option,
            defaults to portage.settings["ROOT"].
    @type target_root: String (Path)
    """
    if vardbapi is None:
        if isinstance(env, config):
            vardbapi = vartree(settings=env).dbapi
        else:
            if target_root is None:
                eprefix = portage.settings["EPREFIX"]
                target_root = portage.settings["ROOT"]
                target_eroot = portage.settings["EROOT"]
            else:
                eprefix = portage.const.EPREFIX
                target_eroot = os.path.join(target_root, eprefix.lstrip(os.sep))
                target_eroot = target_eroot.rstrip(os.sep) + os.sep
            if hasattr(portage, "db") and target_eroot in portage.db:
                vardbapi = portage.db[target_eroot]["vartree"].dbapi
            else:
                settings = config(
                    config_root=target_root, target_root=target_root, eprefix=eprefix
                )
                target_root = settings["ROOT"]
                if env is None:
                    env = settings
                vardbapi = vartree(settings=settings).dbapi

    # Lock the config memory file to prevent symlink creation
    # in merge_contents from overlapping with env-update.
    vardbapi._fs_lock()
    try:
        return _env_update(
            makelinks, target_root, prev_mtimes, contents, env, writemsg_level
        )
    finally:
        vardbapi._fs_unlock()


def _env_update(makelinks, target_root, prev_mtimes, contents, env, writemsg_level):
    if writemsg_level is None:
        writemsg_level = portage.util.writemsg_level
    if target_root is None:
        target_root = portage.settings["ROOT"]
    if prev_mtimes is None:
        prev_mtimes = portage.mtimedb["ldpath"]
    if env is None:
        settings = portage.settings
    else:
        settings = env

    eprefix = settings.get("EPREFIX", "")
    eprefix_lstrip = eprefix.lstrip(os.sep)
    eroot = (
        normalize_path(os.path.join(target_root, eprefix_lstrip)).rstrip(os.sep)
        + os.sep
    )
    envd_dir = os.path.join(eroot, "etc", "env.d")
    ensure_dirs(envd_dir, mode=0o755)
    fns = listdir(envd_dir, EmptyOnError=1)
    fns.sort()
    templist = []
    for x in fns:
        if len(x) < 3:
            continue
        if not x[0].isdigit() or not x[1].isdigit():
            continue
        if x.startswith(".") or x.endswith("~") or x.endswith(".bak"):
            continue
        templist.append(x)
    fns = templist
    del templist

    space_separated = {"CONFIG_PROTECT", "CONFIG_PROTECT_MASK"}
    colon_separated = {
        "ADA_INCLUDE_PATH",
        "ADA_OBJECTS_PATH",
        "CLASSPATH",
        "INFODIR",
        "INFOPATH",
        "KDEDIRS",
        "LDPATH",
        "MANPATH",
        "PATH",
        "PKG_CONFIG_PATH",
        "PRELINK_PATH",
        "PRELINK_PATH_MASK",
        "PYTHONPATH",
        "ROOTPATH",
    }

    config_list = []

    for x in fns:
        file_path = os.path.join(envd_dir, x)
        try:
            myconfig = getconfig(file_path, expand=False)
        except ParseError as e:
            writemsg(f"!!! '{str(e)}'\n", noiselevel=-1)
            del e
            continue
        if myconfig is None:
            # broken symlink or file removed by a concurrent process
            writemsg(f"!!! File Not Found: '{file_path}'\n", noiselevel=-1)
            continue

        config_list.append(myconfig)
        if "SPACE_SEPARATED" in myconfig:
            space_separated.update(myconfig["SPACE_SEPARATED"].split())
            del myconfig["SPACE_SEPARATED"]
        if "COLON_SEPARATED" in myconfig:
            colon_separated.update(myconfig["COLON_SEPARATED"].split())
            del myconfig["COLON_SEPARATED"]

    env = {}
    specials = {}
    for var in space_separated:
        mylist = []
        for myconfig in config_list:
            if var in myconfig:
                for item in myconfig[var].split():
                    if item and not item in mylist:
                        mylist.append(item)
                del myconfig[var]  # prepare for env.update(myconfig)
        if mylist:
            env[var] = " ".join(mylist)
        specials[var] = mylist

    for var in colon_separated:
        mylist = []
        for myconfig in config_list:
            if var in myconfig:
                for item in myconfig[var].split(":"):
                    if item and not item in mylist:
                        mylist.append(item)
                del myconfig[var]  # prepare for env.update(myconfig)
        if mylist:
            env[var] = ":".join(mylist)
        specials[var] = mylist

    for myconfig in config_list:
        """Cumulative variables have already been deleted from myconfig so that
        they won't be overwritten by this dict.update call."""
        env.update(myconfig)

    ldsoconf_path = os.path.join(eroot, "etc", "ld.so.conf")
    try:
        myld = open(
            _unicode_encode(ldsoconf_path, encoding=_encodings["fs"], errors="strict"),
            encoding=_encodings["content"],
            errors="replace",
        )
        myldlines = myld.readlines()
        myld.close()
        oldld = []
        for x in myldlines:
            # each line has at least one char (a newline)
            if x[:1] == "#":
                continue
            oldld.append(x[:-1])
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        oldld = None

    newld = specials["LDPATH"]
    if oldld != newld:
        # ld.so.conf needs updating and ldconfig needs to be run
        myfd = atomic_ofstream(ldsoconf_path)
        myfd.write("# ld.so.conf autogenerated by env-update; make all changes to\n")
        myfd.write("# contents of /etc/env.d directory\n")
        for x in specials["LDPATH"]:
            myfd.write(x + "\n")
        myfd.close()

    potential_lib_dirs = set()
    for lib_dir_glob in ("usr/lib*", "lib*"):
        x = os.path.join(eroot, lib_dir_glob)
        for y in glob.glob(
            _unicode_encode(x, encoding=_encodings["fs"], errors="strict")
        ):
            try:
                y = _unicode_decode(y, encoding=_encodings["fs"], errors="strict")
            except UnicodeDecodeError:
                continue
            if os.path.basename(y) != "libexec":
                potential_lib_dirs.add(y[len(eroot) :])

    # Update prelink.conf if we are prelink-enabled
    if prelink_capable:
        prelink_d = os.path.join(eroot, "etc", "prelink.conf.d")
        ensure_dirs(prelink_d)
        newprelink = atomic_ofstream(os.path.join(prelink_d, "portage.conf"))
        newprelink.write(
            "# prelink.conf autogenerated by env-update; make all changes to\n"
        )
        newprelink.write("# contents of /etc/env.d directory\n")

        for x in sorted(potential_lib_dirs) + ["bin", "sbin"]:
            newprelink.write(f"-l /{x}\n")
        prelink_paths = set()
        prelink_paths |= set(specials.get("LDPATH", []))
        prelink_paths |= set(specials.get("PATH", []))
        prelink_paths |= set(specials.get("PRELINK_PATH", []))
        prelink_path_mask = specials.get("PRELINK_PATH_MASK", [])
        for x in prelink_paths:
            if not x:
                continue
            if x[-1:] != "/":
                x += "/"
            plmasked = 0
            for y in prelink_path_mask:
                if not y:
                    continue
                if y[-1] != "/":
                    y += "/"
                if y == x[0 : len(y)]:
                    plmasked = 1
                    break
            if not plmasked:
                newprelink.write(f"-h {x}\n")
        for x in prelink_path_mask:
            newprelink.write(f"-b {x}\n")
        newprelink.close()

        # Migration code path.  If /etc/prelink.conf was generated by us, then
        # point it to the new stuff until the prelink package re-installs.
        prelink_conf = os.path.join(eroot, "etc", "prelink.conf")
        try:
            with open(
                _unicode_encode(
                    prelink_conf, encoding=_encodings["fs"], errors="strict"
                ),
                "rb",
            ) as f:
                if (
                    f.readline()
                    == b"# prelink.conf autogenerated by env-update; make all changes to\n"
                ):
                    f = atomic_ofstream(prelink_conf)
                    f.write("-c /etc/prelink.conf.d/*.conf\n")
                    f.close()
        except OSError as e:
            if e.errno != errno.ENOENT:
                raise

    current_time = int(time.time())
    mtime_changed = False

    lib_dirs = set()
    for lib_dir in set(specials["LDPATH"]) | potential_lib_dirs:
        x = os.path.join(eroot, lib_dir.lstrip(os.sep))
        try:
            newldpathtime = os.stat(x)[stat.ST_MTIME]
            lib_dirs.add(normalize_path(x))
        except OSError as oe:
            if oe.errno == errno.ENOENT:
                try:
                    del prev_mtimes[x]
                except KeyError:
                    pass
                # ignore this path because it doesn't exist
                continue
            raise
        if newldpathtime == current_time:
            # Reset mtime to avoid the potential ambiguity of times that
            # differ by less than 1 second.
            newldpathtime -= 1
            os.utime(x, (newldpathtime, newldpathtime))
            prev_mtimes[x] = newldpathtime
            mtime_changed = True
        elif x in prev_mtimes:
            if prev_mtimes[x] == newldpathtime:
                pass
            else:
                prev_mtimes[x] = newldpathtime
                mtime_changed = True
        else:
            prev_mtimes[x] = newldpathtime
            mtime_changed = True

    if makelinks and not mtime_changed and contents is not None:
        libdir_contents_changed = False
        for mypath, mydata in contents.items():
            if mydata[0] not in ("obj", "sym"):
                continue
            head, tail = os.path.split(mypath)
            if head in lib_dirs:
                libdir_contents_changed = True
                break
        if not libdir_contents_changed:
            makelinks = False

    if (
        "CHOST" in settings
        and "CBUILD" in settings
        and settings["CHOST"] != settings["CBUILD"]
    ):
        ldconfig = find_binary(f"{settings['CHOST']}-ldconfig")
    else:
        ldconfig = os.path.join(eroot, "sbin", "ldconfig")

    if ldconfig is None:
        pass
    elif not (os.access(ldconfig, os.X_OK) and os.path.isfile(ldconfig)):
        ldconfig = None

    # Only run ldconfig as needed
    if makelinks and ldconfig:
        # ldconfig has very different behaviour between FreeBSD and Linux
        if ostype == "Linux" or ostype.lower().endswith("gnu"):
            # We can't update links if we haven't cleaned other versions first, as
            # an older package installed ON TOP of a newer version will cause ldconfig
            # to overwrite the symlinks we just made. -X means no links. After 'clean'
            # we can safely create links.
            writemsg_level(_(f">>> Regenerating {target_root}etc/ld.so.cache...\n"))
            ret = subprocess.run(
                [ldconfig, "-X", "-r", target_root], cwd="/"
            ).returncode
        elif ostype in ("FreeBSD", "DragonFly"):
            writemsg_level(
                _(f">>> Regenerating {target_root}var/run/ld-elf.so.hints...\n")
            )
            ret = subprocess.run(
                [
                    ldconfig,
                    "-elf",
                    "-i",
                    "-f",
                    f"{target_root}var/run/ld-elf.so.hints",
                    f"{target_root}etc/ld.so.conf",
                ],
                cwd="/",
            ).returncode

        if ret > 0:
            writemsg(f"!!! ldconfig failed with exit status {ret}\n", noiselevel=-1)
        if ret < 0:
            writemsg(f"!!! ldconfig was killed with signal {-ret}\n", noiselevel=-1)

    del specials["LDPATH"]

    notice = "# THIS FILE IS AUTOMATICALLY GENERATED BY env-update.\n"
    notice += "# DO NOT EDIT THIS FILE."
    penvnotice = notice + " CHANGES TO STARTUP PROFILES\n"
    cenvnotice = penvnotice[:]
    penvnotice += "# GO INTO /etc/profile NOT /etc/profile.env\n\n"
    cenvnotice += "# GO INTO /etc/csh.cshrc NOT /etc/csh.env\n\n"

    # create /etc/profile.env for bash support
    profile_env_path = os.path.join(eroot, "etc", "profile.env")
    with atomic_ofstream(profile_env_path) as outfile:
        outfile.write(penvnotice)

        env_keys = [x for x in env if x != "LDPATH"]
        env_keys.sort()
        for k in env_keys:
            v = env[k]
            if v.startswith("$") and not v.startswith("${"):
                outfile.write(f"export {k}=$'{v[1:]}'\n")
            else:
                outfile.write(f"export {k}='{v}'\n")

    # Create the systemd user environment configuration file
    # /etc/environment.d/10-gentoo-env.conf with the
    # environment configuration from /etc/env.d.
    systemd_environment_dir = os.path.join(eroot, "etc", "environment.d")
    os.makedirs(systemd_environment_dir, exist_ok=True)

    systemd_gentoo_env_path = os.path.join(
        systemd_environment_dir, "10-gentoo-env.conf"
    )
    with atomic_ofstream(systemd_gentoo_env_path) as systemd_gentoo_env:
        senvnotice = notice + "\n\n"
        systemd_gentoo_env.write(senvnotice)

        for env_key in env_keys:
            env_key_value = env[env_key]

            # Skip variables with the empty string
            # as value. Those sometimes appear in
            # profile.env (e.g. "export GCC_SPECS=''"),
            # but are invalid in systemd's syntax.
            if not env_key_value:
                continue

            # Transform into systemd environment.d
            # conf syntax, basically shell variable
            # assignment (without "export ").
            line = f"{env_key}={env_key_value}\n"

            systemd_gentoo_env.write(line)

    # create /etc/csh.env for (t)csh support
    outfile = atomic_ofstream(os.path.join(eroot, "etc", "csh.env"))
    outfile.write(cenvnotice)
    for x in env_keys:
        outfile.write(f"setenv {x} '{env[x]}'\n")
    outfile.close()



# Copyright 2018-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from concurrent.futures import Future, ThreadPoolExecutor
import contextlib

import threading

import weakref
import time

import portage
from portage.tests import TestCase
from portage.util._eventloop.global_event_loop import global_event_loop
from portage.util.backoff import RandomExponentialBackoff
from portage.util.futures import asyncio
from portage.util.futures.retry import retry
from portage.util.futures.executor.fork import ForkExecutor


class SucceedLaterException(Exception):
    pass


class SucceedLater:
    """
    A callable object that succeeds some duration of time has passed.
    """

    def __init__(self, duration):
        self._succeed_time = time.monotonic() + duration

    async def __call__(self):
        remaining = self._succeed_time - time.monotonic()
        if remaining > 0:
            await asyncio.sleep(remaining)
        return "success"


class SucceedNeverException(Exception):
    pass


class SucceedNever:
    """
    A callable object that never succeeds.
    """

    async def __call__(self):
        raise SucceedNeverException("expected failure")


class HangForever:
    """
    A callable object that sleeps forever.
    """

    async def __call__(self):
        while True:
            await asyncio.sleep(9)


class RetryTestCase(TestCase):
    @contextlib.contextmanager
    def _wrap_coroutine_func(self, coroutine_func):
        """
        Derived classes may override this method in order to implement
        alternative forms of execution.
        """
        yield coroutine_func

    def testSucceedLater(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedLater(1)) as func_coroutine:
            decorator = retry(
                try_max=9999,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            result = loop.run_until_complete(decorated_func())
            self.assertEqual(result, "success")

    def testSucceedNever(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_max=4,
                try_timeout=None,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, SucceedNeverException)
            )

    def testSucceedNeverReraise(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                reraise=True,
                try_max=4,
                try_timeout=None,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(isinstance(done.pop().exception(), SucceedNeverException))

    def testHangForever(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                try_max=2,
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, asyncio.TimeoutError)
            )

    def testHangForeverReraise(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                reraise=True,
                try_max=2,
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(isinstance(done.pop().exception(), asyncio.TimeoutError))

    def testCancelRetry(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            future = decorated_func()
            loop.call_later(0.3, future.cancel)
            done, pending = loop.run_until_complete(asyncio.wait([future], loop=loop))
            self.assertEqual(len(done), 1)
            self.assertTrue(done.pop().cancelled())

    def testOverallTimeoutWithException(self):
        loop = global_event_loop()
        with self._wrap_coroutine_func(SucceedNever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                overall_timeout=0.3,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            cause = done.pop().exception().__cause__
            self.assertTrue(
                isinstance(
                    cause,
                    (asyncio.TimeoutError, SucceedNeverException),
                ),
                msg=f"Cause was {cause.__class__.__name__}",
            )

    def testOverallTimeoutWithTimeoutError(self):
        loop = global_event_loop()
        # results in TimeoutError because it hangs forever
        with self._wrap_coroutine_func(HangForever()) as func_coroutine:
            decorator = retry(
                try_timeout=0.1,
                overall_timeout=0.3,
                delay_func=RandomExponentialBackoff(multiplier=0.1, base=2),
            )
            decorated_func = decorator(func_coroutine, loop=loop)
            done, pending = loop.run_until_complete(
                asyncio.wait([decorated_func()], loop=loop)
            )
            self.assertEqual(len(done), 1)
            self.assertTrue(
                isinstance(done.pop().exception().__cause__, asyncio.TimeoutError)
            )


class RetryForkExecutorTestCase(RetryTestCase):
    """
    Wrap each coroutine function with AbstractEventLoop.run_in_executor,
    in order to test the event loop's default executor. The executor
    may use either a thread or a subprocess, and either case is
    automatically detected and handled.
    """

    def __init__(self, *pargs, **kwargs):
        super().__init__(*pargs, **kwargs)
        self._executor = None

    def _setUpExecutor(self):
        self._executor = ForkExecutor()

    def _tearDownExecutor(self):
        if self._executor is not None:
            self._executor.shutdown(wait=True)
            self._executor = None

    def setUp(self):
        self._setUpExecutor()

    def tearDown(self):
        self._tearDownExecutor()

    @contextlib.contextmanager
    def _wrap_coroutine_func(self, coroutine_func):
        parent_loop = global_event_loop()
        parent_pid = portage.getpid()
        pending = weakref.WeakValueDictionary()

        # Since ThreadPoolExecutor does not propagate cancellation of a
        # parent_future to the underlying coroutine, use kill_switch to
        # propagate task cancellation to wrapper, so that HangForever's
        # thread returns when retry eventually cancels parent_future.
        def wrapper(kill_switch):
            if portage.getpid() == parent_pid:
                # thread in main process
                def done_callback(result):
                    result.cancelled() or result.exception() or result.result()
                    kill_switch.set()

                def start_coroutine(future):
                    result = asyncio.ensure_future(coroutine_func(), loop=parent_loop)
                    pending[id(result)] = result
                    result.add_done_callback(done_callback)
                    future.set_result(result)

                future = Future()
                parent_loop.call_soon_threadsafe(start_coroutine, future)
                kill_switch.wait()
                if not future.done():
                    future.cancel()
                    raise asyncio.CancelledError
                elif not future.result().done():
                    future.result().cancel()
                    raise asyncio.CancelledError
                else:
                    return future.result().result()

            # child process
            loop = global_event_loop()
            try:
                return loop.run_until_complete(coroutine_func())
            finally:
                loop.close()

        def execute_wrapper():
            kill_switch = threading.Event()
            parent_future = asyncio.ensure_future(
                parent_loop.run_in_executor(self._executor, wrapper, kill_switch),
                loop=parent_loop,
            )

            def kill_callback(parent_future):
                if not kill_switch.is_set():
                    kill_switch.set()

            parent_future.add_done_callback(kill_callback)
            return parent_future

        try:
            yield execute_wrapper
        finally:
            while True:
                try:
                    _, future = pending.popitem()
                except KeyError:
                    break
                try:
                    parent_loop.run_until_complete(future)
                except (Exception, asyncio.CancelledError):
                    pass
                future.cancelled() or future.exception() or future.result()


class RetryThreadExecutorTestCase(RetryForkExecutorTestCase):
    def _setUpExecutor(self):
        self._executor = ThreadPoolExecutor(max_workers=1)



# test_doins.py -- Portage Unit Testing Functionality
# Copyright 2007-2010 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import grp
import os
import pwd
import stat

from portage.tests.bin import setup_env
from portage import tests

doins = setup_env.doins
exists_in_D = setup_env.exists_in_D


class DoIns(setup_env.BinTestCase):
    def testDoIns(self):
        """Tests the most basic senario."""
        self.init()
        try:
            env = setup_env.env
            # Create a file to be installed.
            test_path = os.path.join(env["S"], "test")
            with open(test_path, "w"):
                pass
            doins("test")
            exists_in_D("/test")
            st = os.lstat(env["D"] + "/test")
            # By default, `install`'s permission is 755.
            if stat.S_IMODE(st.st_mode) != 0o755:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsOption(self):
        """Tests with INSOPTIONS doins.py understands."""
        self.init()
        try:
            env = setup_env.env
            env["INSOPTIONS"] = "-pm0644"
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if stat.S_IMODE(st.st_mode) != 0o644:
                raise tests.TestCase.failureException
            self.assertEqual(
                os.stat(os.path.join(env["S"], "test"))[stat.ST_MTIME],
                st[stat.ST_MTIME],
            )
        finally:
            self.cleanup()

    def testDoInsOptionUnsupportedMode(self):
        """Tests with INSOPTIONS in the format doins.py doesn't know."""
        self.init()
        try:
            env = setup_env.env
            # Parse test for -m with unsupported format.
            # It should fall back to `install` command.
            env["INSOPTIONS"] = "-m u+r"
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if stat.S_IMODE(st.st_mode) != 0o400:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsOptionUid(self):
        """Tests setting owner by uid works."""
        self.init()
        try:
            env = setup_env.env
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            uid = os.lstat(os.path.join(env["S"], "test")).st_uid
            # Set owner option with uid. No guarantee that this
            # runs with capability, we set the current UID so that
            # chown should success, although it is difficult to
            # check if chown actually runs or not.
            env["INSOPTIONS"] = "-o %d" % uid
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if st.st_uid != uid:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsOptionUserName(self):
        """Tests setting owner by name works."""
        self.init()
        try:
            env = setup_env.env
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            uid = os.lstat(os.path.join(env["S"], "test")).st_uid
            pw = pwd.getpwuid(uid)
            # Similary to testDoInsOptionUid, use user name.
            env["INSOPTIONS"] = f"-o {pw.pw_name}"
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if st.st_uid != uid:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsOptionGid(self):
        """Tests setting group by gid works."""
        self.init()
        try:
            env = setup_env.env
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            gid = os.lstat(os.path.join(env["S"], "test")).st_gid
            # Similary to testDoInsOptionUid, use gid.
            env["INSOPTIONS"] = "-g %d" % gid
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if st.st_gid != gid:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsOptionGroupName(self):
        """Tests setting group by name works."""
        self.init()
        try:
            env = setup_env.env
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            gid = os.lstat(os.path.join(env["S"], "test")).st_gid
            gr = grp.getgrgid(gid)
            # Similary to testDoInsOptionUid, use group name.
            env["INSOPTIONS"] = f"-g {gr.gr_name}"
            doins("test")
            st = os.lstat(env["D"] + "/test")
            if st.st_gid != gid:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoInsFallback(self):
        """Tests if falling back to the `install` command works."""
        self.init()
        try:
            env = setup_env.env
            # Use an option which doins.py does not know.
            # Then, fallback to `install` command is expected.
            env["INSOPTIONS"] = "-b"
            with open(os.path.join(env["S"], "test"), "w"):
                pass
            doins("test")
            # So `install` should still work.
            exists_in_D("/test")
        finally:
            self.cleanup()

    def testDoInsRecursive(self):
        """Tests installing a directory recursively."""
        self.init()
        try:
            env = setup_env.env
            os.mkdir(os.path.join(env["S"], "testdir"))
            with open(os.path.join(env["S"], "testdir/test"), "w"):
                pass
            doins("-r testdir")
            exists_in_D("/testdir/test")
        finally:
            self.cleanup()

    def testDoDirOption(self):
        """Tests with DIROPTIONS."""
        self.init()
        try:
            env = setup_env.env
            # Use an option which doins.py knows.
            env["DIROPTIONS"] = "-m0755"
            os.mkdir(os.path.join(env["S"], "testdir"))
            with open(os.path.join(env["S"], "testdir/test"), "w"):
                pass
            doins("-r testdir")
            st = os.lstat(env["D"] + "/testdir")
            if stat.S_IMODE(st.st_mode) != 0o755:
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testDoDirFallback(self):
        """Tests with DIROPTIONS which doins.py doesn't understand."""
        self.init()
        try:
            env = setup_env.env
            # Use an option which doins.py does not know.
            # Then, fallback to `install` command is expected.
            env["DIROPTIONS"] = "-p"
            os.mkdir(os.path.join(env["S"], "testdir"))
            with open(os.path.join(env["S"], "testdir/test"), "w"):
                pass
            doins("-r testdir")
            # So, `install` should still work.
            exists_in_D("/testdir/test")
        finally:
            self.cleanup()

    def testSymlinkFile(self):
        """Tests if installing a symlink works.

        In EAPI=4 and later, installing a symlink should creates a
        symlink at destination.
        """
        self.init()
        try:
            env = setup_env.env
            env["EAPI"] = "4"  # Enable symlink.
            env["ED"] = env["D"]
            env["PORTAGE_ACTUAL_DISTDIR"] = "/foo"
            # Create a file to be installed.
            test_path = os.path.join(env["S"], "test")
            with open(test_path, "w"):
                pass
            symlink_path = os.path.join(env["S"], "symlink")
            os.symlink("test", symlink_path)
            doins("test symlink")
            exists_in_D("/symlink")
            # Make sure installed symlink is actually a symbolic
            # link pointing to test.
            if not os.path.islink(env["D"] + "/symlink"):
                raise tests.TestCase.failureException
            if os.readlink(env["D"] + "/symlink") != "test":
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testSymlinkFileRecursive(self):
        """Tests if installing a symlink in a directory works."""
        self.init()
        try:
            env = setup_env.env
            env["EAPI"] = "4"  # Enable symlink.
            env["ED"] = env["D"]
            env["PORTAGE_ACTUAL_DISTDIR"] = "/foo"
            # Create a file to be installed.
            parent_path = os.path.join(env["S"], "test")
            os.mkdir(parent_path)
            with open(os.path.join(parent_path, "test"), "w"):
                pass
            symlink_path = os.path.join(env["S"], "test", "symlink")
            os.symlink("test", symlink_path)
            doins("-r test")
            exists_in_D("/test/symlink")
            # Make sure installed symlink is actually a symbolic
            # link pointing to test.
            if not os.path.islink(env["D"] + "/test/symlink"):
                raise tests.TestCase.failureException
            if os.readlink(env["D"] + "/test/symlink") != "test":
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testSymlinkDir(self):
        """Tests installing a symlink to a directory."""
        self.init()
        try:
            env = setup_env.env
            env["EAPI"] = "4"  # Enable symlink.
            env["ED"] = env["D"]
            env["PORTAGE_ACTUAL_DISTDIR"] = "/foo"
            # Create a dir to be installed.
            os.mkdir(os.path.join(env["S"], "test"))
            symlink_path = os.path.join(env["S"], "symlink")
            os.symlink("test", symlink_path)
            doins("test symlink")
            # Make sure installed symlink is actually a symbolic
            # link pointing to test.
            if not os.path.islink(env["D"] + "/symlink"):
                raise tests.TestCase.failureException
            if os.readlink(env["D"] + "/symlink") != "test":
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testSymlinkDirRecursive(self):
        """Tests installing a symlink to a directory under a directory."""
        self.init()
        try:
            env = setup_env.env
            env["EAPI"] = "4"  # Enable symlink.
            env["ED"] = env["D"]
            env["PORTAGE_ACTUAL_DISTDIR"] = "/foo"
            # Create a file to be installed.
            parent_path = os.path.join(env["S"], "test")
            os.mkdir(parent_path)
            os.mkdir(os.path.join(parent_path, "test"))
            symlink_path = os.path.join(env["S"], "test", "symlink")
            os.symlink("test", symlink_path)
            doins("-r test")
            # Make sure installed symlink is actually a symbolic
            # link pointing to test.
            if not os.path.islink(env["D"] + "/test/symlink"):
                raise tests.TestCase.failureException
            if not os.path.isdir(env["D"] + "/test/symlink"):
                raise tests.TestCase.failureException
            if os.readlink(env["D"] + "/test/symlink") != "test":
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testSymlinkOverwrite(self):
        """Tests installing a file to overwrite the existing file.

        Specifically, if the existing file is a symlink, it should be
        removed once, and the file content should be copied.
        """
        self.init()
        try:
            env = setup_env.env
            test_path = os.path.join(env["S"], "test")
            with open(test_path, "w"):
                pass
            # Create a dangling symlink. If removal does not work,
            # this would easily cause ENOENT error.
            os.symlink("foo/bar", env["D"] + "/test")
            doins("test")
            # Actual file should be installed.
            if os.path.islink(env["D"] + "/test"):
                raise tests.TestCase.failureException
        finally:
            self.cleanup()

    def testHardlinkOverwrite(self):
        """Tests installing a file to overwrite the hardlink.

        If the existing file is a hardlink, it should be removed once,
        and the file content should be copied.
        """
        self.init()
        try:
            env = setup_env.env
            test_path = os.path.join(env["S"], "test")
            with open(test_path, "w"):
                pass
            # Create hardlink at the dest.
            os.link(test_path, env["D"] + "/test")
            doins("test")
            # The hardlink should be unlinked, and then a copy
            # should be created.
            if os.path.samefile(test_path, env["D"] + "/test"):
                raise tests.TestCase.failureException
        finally:
            self.cleanup()



# Copyright 2016-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = [
    "ConfigParserError",
    "NoOptionError",
    "ParsingError",
    "RawConfigParser",
    "SafeConfigParser",
    "read_configs",
]

# the following scary compatibility thing provides two classes:
# - SafeConfigParser that provides safe interpolation for values,
# - RawConfigParser that provides no interpolation for values.

import io

from configparser import (
    Error as ConfigParserError,
    NoOptionError,
    ParsingError,
    RawConfigParser,
)
from configparser import ConfigParser as SafeConfigParser

from portage import _encodings
from portage import _unicode_encode


def read_configs(parser, paths):
    """
    Read configuration files from given paths into the specified
    ConfigParser, handling path encoding portably.
    @param parser: target *ConfigParser instance
    @type parser: SafeConfigParser or RawConfigParser
    @param paths: list of paths to read
    @type paths: iterable
    """
    # use read_file/readfp in order to control decoding of unicode
    try:
        # Python >=3.2
        read_file = parser.read_file
        source_kwarg = "source"
    except AttributeError:
        read_file = parser.readfp
        source_kwarg = "filename"

    for p in paths:
        if isinstance(p, str):
            f = None
            try:
                f = open(
                    _unicode_encode(p, encoding=_encodings["fs"], errors="strict"),
                    encoding=_encodings["repo.content"],
                    errors="replace",
                )
            except OSError:
                pass
            else:
                # The 'source' keyword argument is needed since otherwise
                # ConfigParser in Python <3.3.3 may throw a TypeError
                # because it assumes that f.name is a native string rather
                # than binary when constructing error messages.
                kwargs = {source_kwarg: p}
                read_file(f, **kwargs)
            finally:
                if f is not None:
                    f.close()
        elif isinstance(p, io.StringIO):
            kwargs = {source_kwarg: "<io.StringIO>"}
            read_file(p, **kwargs)
        else:
            raise TypeError(
                f"Unsupported type {type(p)!r} of element {p!r} of 'paths' argument"
            )



# Copyright 1999-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import collections
from itertools import chain

from portage.output import red
from portage.util import cmp_sort_key
from portage.output import blue


class UseFlagDisplay:
    __slots__ = ("name", "enabled", "forced")

    def __init__(self, name, enabled, forced):
        self.name = name
        self.enabled = enabled
        self.forced = forced

    def __str__(self):
        s = self.name
        if self.enabled:
            s = red(s)
        else:
            s = "-" + s
            s = blue(s)
        if self.forced:
            s = f"({s})"
        return s

    def _cmp_combined(a, b):
        """
        Sort by name, combining enabled and disabled flags.
        """
        return (a.name > b.name) - (a.name < b.name)

    sort_combined = cmp_sort_key(_cmp_combined)
    del _cmp_combined

    def _cmp_separated(a, b):
        """
        Sort by name, separating enabled flags from disabled flags.
        """
        enabled_diff = b.enabled - a.enabled
        if enabled_diff:
            return enabled_diff
        return (a.name > b.name) - (a.name < b.name)

    sort_separated = cmp_sort_key(_cmp_separated)
    del _cmp_separated


_flag_info = collections.namedtuple("_flag_info", ("flag", "display"))


def pkg_use_display(pkg, opts, modified_use=None):
    settings = pkg.root_config.settings
    use_expand = pkg.use.expand
    use_expand_hidden = pkg.use.expand_hidden
    alphabetical_use = "--alphabetical" in opts
    forced_flags = set(chain(pkg.use.force, pkg.use.mask))
    if modified_use is None:
        use = set(pkg.use.enabled)
    else:
        use = set(modified_use)
    use.discard(settings.get("ARCH"))
    use_expand_flags = set()
    use_enabled = {}
    use_disabled = {}
    for varname in use_expand:
        flag_prefix = varname.lower() + "_"
        for f in use:
            if f.startswith(flag_prefix):
                use_expand_flags.add(f)
                use_enabled.setdefault(varname.upper(), []).append(
                    _flag_info(f, f[len(flag_prefix) :])
                )

        for f in pkg.iuse.all:
            if f.startswith(flag_prefix):
                use_expand_flags.add(f)
                if f not in use:
                    use_disabled.setdefault(varname.upper(), []).append(
                        _flag_info(f, f[len(flag_prefix) :])
                    )

    var_order = set(use_enabled)
    var_order.update(use_disabled)
    var_order = sorted(var_order)
    var_order.insert(0, "USE")
    use.difference_update(use_expand_flags)
    use_enabled["USE"] = list(_flag_info(f, f) for f in use)
    use_disabled["USE"] = []

    for f in pkg.iuse.all:
        if f not in use and f not in use_expand_flags:
            use_disabled["USE"].append(_flag_info(f, f))

    flag_displays = []
    for varname in var_order:
        if varname.lower() in use_expand_hidden:
            continue
        flags = []
        for f in use_enabled.get(varname, []):
            flags.append(UseFlagDisplay(f.display, True, f.flag in forced_flags))
        for f in use_disabled.get(varname, []):
            flags.append(UseFlagDisplay(f.display, False, f.flag in forced_flags))
        if alphabetical_use:
            flags.sort(key=UseFlagDisplay.sort_combined)
        else:
            flags.sort(key=UseFlagDisplay.sort_separated)
        flag_displays.append(f"{varname}=\"{' '.join(f'{f}' for f in flags)}\"")

    return " ".join(flag_displays)



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.AbstractDepPriority import AbstractDepPriority


class UnmergeDepPriority(AbstractDepPriority):
    __slots__ = (
        "ignored",
        "optional",
        "satisfied",
    )
    """
	Combination of properties           Priority  Category

	runtime_slot_op                        0       HARD
	runtime                               -1       HARD
	runtime_post                          -2       HARD
	buildtime                             -3       SOFT
	(none of the above)                   -3       SOFT
	"""

    MAX = 0
    SOFT = -3
    MIN = -3

    def __init__(self, **kwargs):
        AbstractDepPriority.__init__(self, **kwargs)
        if self.buildtime:
            self.optional = True

    def __int__(self):
        if self.runtime_slot_op:
            return 0
        if self.runtime:
            return -1
        if self.runtime_post:
            return -2
        if self.buildtime:
            return -3
        return -3

    def __str__(self):
        if self.ignored:
            return "ignored"
        if self.runtime_slot_op:
            return "hard slot op"
        myvalue = self.__int__()
        if myvalue > self.SOFT:
            return "hard"
        return "soft"



# Copyright 1999-2011 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from _emerge.Task import Task


class Blocker(Task):
    __hash__ = Task.__hash__
    __slots__ = ("root", "atom", "cp", "eapi", "priority", "satisfied")

    def __init__(self, **kwargs):
        Task.__init__(self, **kwargs)
        self.cp = self.atom.cp
        self._hash_key = ("blocks", self.root, self.atom, self.eapi)
        self._hash_value = hash(self._hash_key)



# Copyright 2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SimpleDepcleanTestCase(TestCase):
    def testSimpleDepclean(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/B:=",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/A",
            },
            "dev-libs/C-1": {},
        }

        installed = {
            "dev-libs/A-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/B:0/0=",
            },
            "dev-libs/B-1": {
                "EAPI": "5",
                "RDEPEND": "dev-libs/A",
            },
            "dev-libs/C-1": {},
        }

        world = ("dev-libs/C",)

        test_cases = (
            # Remove dev-libs/A-1 first because of dev-libs/B:0/0= (built
            # slot-operator dep).
            ResolverPlaygroundTestCase(
                [],
                options={"--depclean": True},
                success=True,
                ordered=True,
                cleanlist=["dev-libs/A-1", "dev-libs/B-1"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 1999-2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import functools
import io
import stat
import textwrap
from _emerge.SpawnProcess import SpawnProcess
from _emerge.EbuildBuildDir import EbuildBuildDir
from _emerge.EbuildIpcDaemon import EbuildIpcDaemon
import portage
from portage.elog import messages as elog_messages
from portage import installation
from portage.package.ebuild._ipc.ExitCommand import ExitCommand
from portage.package.ebuild._ipc.QueryCommand import QueryCommand
from portage import os
from portage.util.futures import asyncio
from portage.util import apply_secpass_permissions, no_color

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.package.ebuild.doebuild:_global_pid_phases",
)


class AbstractEbuildProcess(SpawnProcess):
    __slots__ = (
        "phase",
        "settings",
        "_build_dir",
        "_build_dir_unlock",
        "_ipc_daemon",
        "_exit_command",
        "_exit_timeout_id",
        "_start_future",
    )

    _phases_without_builddir = (
        "clean",
        "cleanrm",
        "depend",
        "help",
    )
    _phases_interactive_whitelist = ("config",)

    # Number of milliseconds to allow natural exit of the ebuild
    # process after it has called the exit command via IPC. It
    # doesn't hurt to be generous here since the scheduler
    # continues to process events during this period, and it can
    # return long before the timeout expires.
    _exit_timeout = 10  # seconds

    # The EbuildIpcDaemon support is well tested, but this variable
    # is left so we can temporarily disable it if any issues arise.
    _enable_ipc_daemon = (
        installation.TYPE == installation.TYPES.SOURCE or "@IPC@" == "True"
    )

    def __init__(self, **kwargs):
        SpawnProcess.__init__(self, **kwargs)
        if self.phase is None:
            phase = self.settings.get("EBUILD_PHASE")
            if not phase:
                phase = "other"
            self.phase = phase

    def _start(self):
        need_builddir = self.phase not in self._phases_without_builddir

        # This can happen if the pre-clean phase triggers
        # die_hooks for some reason, and PORTAGE_BUILDDIR
        # doesn't exist yet.
        if need_builddir and not os.path.isdir(self.settings["PORTAGE_BUILDDIR"]):
            msg = (
                f"The ebuild phase '{self.phase}' has been aborted since "
                f"PORTAGE_BUILDDIR does not exist: '{self.settings['PORTAGE_BUILDDIR']}'"
            )
            self._eerror(textwrap.wrap(msg, 72))
            self.returncode = 1
            self._async_wait()
            return

        if self.background:
            # Automatically prevent color codes from showing up in logs,
            # since we're not displaying to a terminal anyway.
            self.settings["NOCOLOR"] = "true"
            self.settings["NO_COLOR"] = "true"

        start_ipc_daemon = False
        if self._enable_ipc_daemon:
            self.settings.pop("PORTAGE_EBUILD_EXIT_FILE", None)
            if self.phase not in self._phases_without_builddir:
                start_ipc_daemon = True
                if "PORTAGE_BUILDDIR_LOCKED" not in self.settings:
                    self._build_dir = EbuildBuildDir(
                        scheduler=self.scheduler, settings=self.settings
                    )
                    self._start_future = self._build_dir.async_lock()
                    self._start_future.add_done_callback(
                        functools.partial(
                            self._start_post_builddir_lock,
                            start_ipc_daemon=start_ipc_daemon,
                        )
                    )
                    return
            else:
                self.settings.pop("PORTAGE_IPC_DAEMON", None)
        else:
            # Since the IPC daemon is disabled, use a simple tempfile based
            # approach to detect unexpected exit like in bug #190128.
            self.settings.pop("PORTAGE_IPC_DAEMON", None)
            if self.phase not in self._phases_without_builddir:
                exit_file = os.path.join(
                    self.settings["PORTAGE_BUILDDIR"], ".exit_status"
                )
                self.settings["PORTAGE_EBUILD_EXIT_FILE"] = exit_file
                try:
                    os.unlink(exit_file)
                except OSError:
                    if os.path.exists(exit_file):
                        # make sure it doesn't exist
                        raise
            else:
                self.settings.pop("PORTAGE_EBUILD_EXIT_FILE", None)

        self._start_post_builddir_lock(start_ipc_daemon=start_ipc_daemon)

    def _start_post_builddir_lock(self, lock_future=None, start_ipc_daemon=False):
        if lock_future is not None:
            if lock_future is not self._start_future:
                raise AssertionError("lock_future is not self._start_future")
            self._start_future = None
            if lock_future.cancelled():
                self._build_dir = None
                self.cancelled = True
                self._was_cancelled()
                self._async_wait()
                return

            lock_future.result()

        if start_ipc_daemon:
            self.settings["PORTAGE_IPC_DAEMON"] = "1"
            self._start_ipc_daemon()

        if self.fd_pipes is None:
            self.fd_pipes = {}
        null_fd = None
        if (
            0 not in self.fd_pipes
            and self.phase not in self._phases_interactive_whitelist
            and "interactive" not in self.settings.get("PROPERTIES", "").split()
        ):
            null_fd = os.open("/dev/null", os.O_RDONLY)
            self.fd_pipes[0] = null_fd

        self.log_filter_file = self.settings.get("PORTAGE_LOG_FILTER_FILE_CMD")
        try:
            SpawnProcess._start(self)
        finally:
            if null_fd is not None:
                os.close(null_fd)

    def _init_ipc_fifos(self):
        input_fifo = os.path.join(self.settings["PORTAGE_BUILDDIR"], ".ipc", "in")
        output_fifo = os.path.join(self.settings["PORTAGE_BUILDDIR"], ".ipc", "out")

        for p in (input_fifo, output_fifo):
            st = None
            try:
                st = os.lstat(p)
            except OSError:
                os.mkfifo(p)
            else:
                if not stat.S_ISFIFO(st.st_mode):
                    st = None
                    try:
                        os.unlink(p)
                    except OSError:
                        pass
                    os.mkfifo(p)

            apply_secpass_permissions(
                p,
                uid=os.getuid(),
                gid=portage.data.portage_gid,
                mode=0o770,
                stat_cached=st,
            )

        return (input_fifo, output_fifo)

    def _start_ipc_daemon(self):
        self._exit_command = ExitCommand()
        self._exit_command.reply_hook = self._exit_command_callback
        query_command = QueryCommand(self.settings, self.phase)
        commands = {
            "available_eclasses": query_command,
            "best_version": query_command,
            "eclass_path": query_command,
            "exit": self._exit_command,
            "has_version": query_command,
            "license_path": query_command,
            "master_repositories": query_command,
            "repository_path": query_command,
        }
        input_fifo, output_fifo = self._init_ipc_fifos()
        self._ipc_daemon = EbuildIpcDaemon(
            commands=commands,
            input_fifo=input_fifo,
            output_fifo=output_fifo,
            scheduler=self.scheduler,
        )
        self._ipc_daemon.start()

    def _exit_command_callback(self):
        if self._registered:
            # Let the process exit naturally, if possible.
            self._exit_timeout_id = self.scheduler.call_later(
                self._exit_timeout, self._exit_command_timeout_cb
            )

    def _exit_command_timeout_cb(self):
        if self._registered:
            # If it doesn't exit naturally in a reasonable amount
            # of time, kill it (solves bug #278895). We try to avoid
            # this when possible since it makes sandbox complain about
            # being killed by a signal.
            self.cancel()
            self._exit_timeout_id = self.scheduler.call_later(
                self._cancel_timeout, self._cancel_timeout_cb
            )
        else:
            self._exit_timeout_id = None

    def _cancel_timeout_cb(self):
        self._exit_timeout_id = None
        self._async_waitpid()

    def _orphan_process_warn(self):
        phase = self.phase

        msg = (
            f"The ebuild phase '{phase}' with pid {self.pid} appears "
            "to have left an orphan process running in the background."
        )

        self._eerror(textwrap.wrap(msg, 72))

    def _can_log(self, slave_fd):
        # With sesandbox, logging works through a pty but not through a
        # normal pipe. So, disable logging if ptys are broken.
        # See Bug #162404.
        # TODO: Add support for logging via named pipe (fifo) with
        # sesandbox, since EbuildIpcDaemon uses a fifo and it's known
        # to be compatible with sesandbox.
        return not (
            "sesandbox" in self.settings.features and self.settings.selinux_enabled()
        ) or os.isatty(slave_fd)

    def _killed_by_signal(self, signum):
        msg = f"The ebuild phase '{self.phase}' has been killed by signal {signum}."
        self._eerror(textwrap.wrap(msg, 72))

    def _unexpected_exit(self):
        phase = self.phase

        msg = (
            f"The ebuild phase '{phase}' has exited "
            "unexpectedly. This type of behavior "
            "is known to be triggered "
            "by things such as failed variable "
            "assignments (bug #190128) or bad substitution "
            "errors (bug #200313). Normally, before exiting, bash should "
            "have displayed an error message above. If bash did not "
            "produce an error message above, it's possible "
            "that the ebuild has called `exit` when it "
            "should have called `die` instead. This behavior may also "
            "be triggered by a corrupt bash binary or a hardware "
            "problem such as memory or cpu malfunction. If the problem is not "
            "reproducible or it appears to occur randomly, then it is likely "
            "to be triggered by a hardware problem. "
            "If you suspect a hardware problem then you should "
            "try some basic hardware diagnostics such as memtest. "
            "Please do not report this as a bug unless it is consistently "
            "reproducible and you are sure that your bash binary and hardware "
            "are functioning properly."
        )

        self._eerror(textwrap.wrap(msg, 72))

    def _eerror(self, lines):
        self._elog("eerror", lines)

    def _elog(self, elog_funcname, lines):
        out = io.StringIO()
        phase = self.phase
        elog_func = getattr(elog_messages, elog_funcname)
        global_havecolor = portage.output.havecolor
        try:
            portage.output.havecolor = not no_color(self.settings)
            for line in lines:
                elog_func(line, phase=phase, key=self.settings.mycpv, out=out)
        finally:
            portage.output.havecolor = global_havecolor
        msg = out.getvalue()
        if msg:
            log_path = None
            if self.settings.get("PORTAGE_BACKGROUND") != "subprocess":
                log_path = self.settings.get("PORTAGE_LOG_FILE")
            self.scheduler.output(msg, log_path=log_path)

    def _async_waitpid_cb(self, *args, **kwargs):
        """
        Override _async_waitpid_cb to perform cleanup that is
        not necessarily idempotent.
        """
        SpawnProcess._async_waitpid_cb(self, *args, **kwargs)

        if self._exit_timeout_id is not None:
            self._exit_timeout_id.cancel()
            self._exit_timeout_id = None

        if self._ipc_daemon is not None:
            self._ipc_daemon.cancel()
            if self._exit_command.exitcode is not None:
                self.returncode = self._exit_command.exitcode
            else:
                if self.returncode < 0:
                    if not self.cancelled:
                        self._killed_by_signal(-self.returncode)
                else:
                    self.returncode = 1
                    if not self.cancelled:
                        self._unexpected_exit()

        elif not self.cancelled:
            exit_file = self.settings.get("PORTAGE_EBUILD_EXIT_FILE")
            if exit_file and not os.path.exists(exit_file):
                if self.returncode < 0:
                    if not self.cancelled:
                        self._killed_by_signal(-self.returncode)
                else:
                    self.returncode = 1
                    if not self.cancelled:
                        self._unexpected_exit()

    def _async_wait(self):
        """
        Override _async_wait to asynchronously unlock self._build_dir
        when necessary.
        """
        if self._build_dir is None:
            SpawnProcess._async_wait(self)
        elif self._build_dir_unlock is None:
            if self.returncode is None:
                raise asyncio.InvalidStateError(f"Result is not ready for {self}")
            self._async_unlock_builddir(returncode=self.returncode)

    def _async_unlock_builddir(self, returncode=None):
        """
        Release the lock asynchronously, and if a returncode parameter
        is given then set self.returncode and notify exit listeners.
        """
        if self._build_dir_unlock is not None:
            raise AssertionError("unlock already in progress")
        if returncode is not None:
            # The returncode will be set after unlock is complete.
            self.returncode = None
        self._build_dir_unlock = self._build_dir.async_unlock()
        # Unlock only once.
        self._build_dir = None
        self._build_dir_unlock.add_done_callback(
            functools.partial(self._unlock_builddir_exit, returncode=returncode)
        )

    def _unlock_builddir_exit(self, unlock_future, returncode=None):
        # Normally, async_unlock should not raise an exception here.
        unlock_future.cancelled() or unlock_future.result()
        if returncode is not None:
            if unlock_future.cancelled():
                self.cancelled = True
                self._was_cancelled()
            else:
                self.returncode = returncode
            SpawnProcess._async_wait(self)



# Copyright 2017 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.dep import Atom, use_reduce
from portage.dep.dep_check import _overlap_dnf


class OverlapDNFTestCase(TestCase):
    def testOverlapDNF(self):
        test_cases = (
            (
                "|| ( cat/A cat/B ) cat/E || ( cat/C cat/D )",
                [["||", "cat/A", "cat/B"], "cat/E", ["||", "cat/C", "cat/D"]],
            ),
            (
                "|| ( cat/A cat/B ) cat/D || ( cat/B cat/C )",
                [
                    "cat/D",
                    [
                        "||",
                        ["cat/A", "cat/B"],
                        ["cat/A", "cat/C"],
                        ["cat/B", "cat/B"],
                        ["cat/B", "cat/C"],
                    ],
                ],
            ),
            (
                "|| ( cat/A cat/B ) || ( cat/C cat/D )  || ( ( cat/B cat/E ) cat/F )",
                [
                    [
                        "||",
                        ["cat/A", "cat/B", "cat/E"],
                        ["cat/A", "cat/F"],
                        ["cat/B", "cat/B", "cat/E"],
                        ["cat/B", "cat/F"],
                    ],
                    ["||", "cat/C", "cat/D"],
                ],
            ),
        )

        for dep_str, result in test_cases:
            self.assertEqual(
                _overlap_dnf(use_reduce(dep_str, token_class=Atom, opconvert=True)),
                result,
            )
